/* alias.c - Command aliases
 * Copyright 2000-2001 srvx Development Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: alias.c,v 1.15.2.2.2.4 2002/05/05 00:40:00 entrope Exp $
 */

#include "config.h"
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "common.h"
#include "dict.h"
#include "recdb.h"
#include "helpfile.h"
#include "alias.h"
#include "log.h"

#define ALIAS_DATABASE_NAME		"aliases.db"
#define ALIAS_TEMPORARY_NAME		"aliases.db.new"

dict_t services;

static void
alias_free_alias(void *data)
{
    struct aliasNode *aNode = data;
    free(aNode->alias);
    free(aNode->command);
    free(aNode);
}

static void
alias_free_command(void *data)
{
    struct commandNode *cNode = data;
    free(cNode->command);
    dict_delete(cNode->aliases);
    free(cNode);
}

int
alias_register_service(const char *service, command_search_func_t search_func, alias_command_func_t alias_func, delete_command_func_t delete_func)
{
    struct serviceNode *sNode = NULL;

    if(!service)
    {
	return 0;
    }

    /* Search for any existing service records and skip out if we
       find one. */
    if((sNode = dict_find(services, service, NULL)))
    {
	return 1;
    }

    sNode = malloc(sizeof(struct serviceNode));
    if(!sNode)
    {
	return 0;
    }

    sNode->service = strdup(service);

    sNode->aliases = dict_new();
    dict_set_free_data(sNode->aliases, alias_free_alias);
    sNode->commands = dict_new();
    dict_set_free_data(sNode->commands, alias_free_command);

    sNode->search_func = search_func;
    sNode->alias_func = alias_func;
    sNode->delete_func = delete_func;

    dict_insert(services, sNode->service, sNode);

    return 1;
}

static void
link_command(struct serviceNode *sNode, struct aliasNode *aNode)
{
    struct commandNode *cNode;

    if(!(cNode = dict_find(sNode->commands, aNode->command, NULL)))
    {
	/* Copy command again so the cNode doesn't get lost when this
	   alias is deleted. */
	cNode = malloc(sizeof(struct commandNode));

	cNode->command = strdup(aNode->command);
	cNode->aliases = dict_new();

	dict_insert(sNode->commands, cNode->command, cNode);
    }

    dict_insert(cNode->aliases, aNode->alias, aNode);
}

static void
unlink_command(struct serviceNode *sNode, struct aliasNode *aNode)
{
    struct commandNode *cNode;

    cNode = dict_find(sNode->commands, aNode->command, NULL);
    if(cNode)
    {
	dict_remove(cNode->aliases, aNode->alias);

	if(!dict_size(cNode->aliases))
	{
	    dict_remove(sNode->commands, aNode->command);
	}
    }
}

int
add_alias(const char *service, const char *alias, char *command)
{
    struct aliasNode *aNode = NULL;
    struct serviceNode *sNode = NULL;
    char *alias_dup;

    if(!service || !alias || !command)
    {
	return 0;
    }

    /* Resolve the alias to avoid creating an alias to an alias. */
    command = resolve_alias(service, (char *)command);

    if(!strcmp(alias, command))
    {
	return ALIAS_SELF;
    }

    if(!(sNode = dict_find(services, service, NULL)))
    {
	return ALIAS_SERVICE_NOT_FOUND;
    }

    aNode = dict_find(sNode->aliases, alias, NULL);

    if(sNode->search_func(alias) && !aNode)
    {
	return ALIAS_FOUND_COMMAND;
    }
    if(!sNode->search_func(command))
    {
	return ALIAS_COMMAND_NOT_FOUND;
    }

    alias_dup = strdup(alias);

    if(!sNode->alias_func(alias_dup, command))
    {
	free(alias_dup);
	return 0;
    }

    if(aNode)
    {
	/* Reuse the aNode and update command only, since the alias name
	   is the same. */
	unlink_command(sNode, aNode);

	free(aNode->command);
	aNode->command = strdup(command);

	link_command(sNode, aNode);

	free(alias_dup);
	return ALIAS_MODIFIED;
    }

    aNode = malloc(sizeof(struct aliasNode));
    if(!aNode)
    {
	return 0;
    }

    aNode->alias = alias_dup;
    aNode->command = strdup(command);

    dict_insert(sNode->aliases, aNode->alias, aNode);
    link_command(sNode, aNode);

    return ALIAS_CREATED;
}

int
del_alias(const char *service, const char *alias)
{
    struct serviceNode *sNode = NULL;
    struct aliasNode *aNode = NULL;

    if(!service || !alias)
    {
	return 0;
    }

    if(!(sNode = dict_find(services, service, NULL)))
    {
	return ALIAS_SERVICE_NOT_FOUND;
    }

    aNode = dict_find(sNode->aliases, alias, NULL);

    if(sNode->search_func(alias) && !aNode)
    {
	return ALIAS_FOUND_COMMAND;
    }
    else if(!aNode)
    {
	return ALIAS_NOT_FOUND;
    }

    sNode->delete_func(alias);
    unlink_command(sNode, aNode);
    dict_remove(sNode->aliases, aNode->alias);
    return ALIAS_DELETED;
}

char*
resolve_alias(const char *service, char *alias)
{
    struct serviceNode *sNode = NULL;
    struct aliasNode *aNode = NULL;

    if(!service || !alias || !(sNode = dict_find(services, service, NULL)))
    {
	return alias;
    }
    
    aNode = dict_find(sNode->aliases, alias, NULL);
    return aNode ? aNode->command : alias;
}

struct aliasList
{
    char *aliases;

    int pos;
    int len;
};

static int
alias_list_helper(const char *key, void *data, void *extra)
{
    struct aliasList *list = extra;
    int len = strlen(key);
    (void)data;

    if(list->pos + len >= list->len)
    {
	return 0;
    }

    (void)memcpy(list->aliases + list->pos, key, len);
    list->pos += len;

    list->aliases[list->pos++] = ',';
    list->aliases[list->pos++] = ' ';

    return 0;
}

char*
command_aliases(const char *service, char *command)
{
    struct serviceNode *sNode = NULL;
    struct commandNode *cNode = NULL;
    struct aliasList list;
    char aliases[MIN_LINE_SIZE * 2];

    if(!service || !command || !(sNode = dict_find(services, service, NULL)))
    {
	return NULL;
    }

    cNode = dict_find(sNode->commands, command, NULL);

    if(!cNode)
    {
	return NULL;
    }

    list.aliases = aliases;
    list.pos = 0;
    list.len = sizeof(aliases);

    dict_foreach(cNode->aliases, alias_list_helper, &list);
    aliases[list.pos - 2] = 0;

    return strdup(aliases);
}

char*
cNode_aliases(struct commandNode *cNode)
{
    struct aliasList list;
    char aliases[MIN_LINE_SIZE * 2];

    if(!cNode)
    {
	return NULL;
    }

    list.aliases = aliases;
    list.pos = 0;
    list.len = sizeof(aliases);

    dict_foreach(cNode->aliases, alias_list_helper, &list);
    aliases[list.pos - 2] = 0;

    return strdup(aliases);
}

void
read_aliases(void)
{
    struct record_data *hir, *rd;
    dict_t db;
    dict_iterator_t it, it2;

    db = parse_database(ALIAS_DATABASE_NAME);
    if(!db) return;
    for(it=dict_first(db); it; it=iter_next(it))
    {
        hir = iter_data(it);
        if(hir->type != RECDB_OBJECT) continue;
        if(alias_register_service(iter_key(it), NULL, NULL, NULL))
        {
            for(it2=dict_first(hir->d.object); it2; it2=iter_next(it2))
            {
                rd = iter_data(it2);
                if(rd->type != RECDB_QSTRING) continue;
                add_alias(iter_key(it), iter_key(it2), rd->d.qstring);
            }
        }
    }
    free_database(db);
}

int
write_alias(const char *key, void *data, void *extra)
{
    struct aliasNode *aNode = data;
    struct record_data *rd;
    (void)key;

    rd = alloc_record_data_qstring(aNode->command);
    dict_insert(extra, aNode->alias, rd);

    return 0;
}

int
write_service(const char *key, void *data, void *extra)
{
    struct serviceNode *sNode = data;
    struct record_data *rd;
    dict_t aliases;

    aliases = alloc_object();

    dict_foreach(sNode->aliases, write_alias, aliases);

    rd = alloc_record_data_object(aliases);
    write_record(extra, key, rd);
    free_record_data(rd);

    return 0;
}

static void
write_aliases(void)
{
    FILE *db = fopen(ALIAS_TEMPORARY_NAME, "w");
    if (!db) return;
    dict_foreach(services, write_service, db);
    fclose(db);

    if (rename(ALIAS_TEMPORARY_NAME, ALIAS_DATABASE_NAME) == -1)
    {
	log(DEBUG_LOG, LOG_ERROR, "Unable to rename new db to old: %s\n", strerror(errno));
    }
}

static void
alias_free_service(void *data)
{
    struct serviceNode *sNode = data;
    dict_delete(sNode->aliases);
    dict_delete(sNode->commands);
    free(sNode);
}

static void
alias_cleanup(void)
{
    /* TODO: implement alias_cleanup */
    dict_delete(services);
}

void
alias_init(void)
{
    services = dict_new();
    dict_set_free_keys(services, free);
    dict_set_free_data(services, alias_free_service);
    reg_db_write_func(write_aliases);
    reg_exit_func(alias_cleanup);
}
