/* gline.c - Gline database
 * Copyright 2000-2001 srvx Development Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: gline.c,v 1.6.2.10.2.7 2002/05/06 01:36:56 entrope Exp $
 */

#include "config.h"
#include <errno.h>
#include <stdlib.h>

#include "common.h"
#include "heap.h"
#include "helpfile.h"
#include "log.h"
#include "proto.h"
#include "recdb.h"
#include "timeq.h"
#include "gline.h"

#define GLINE_DB_NAME "gline.db"
#define TEMP_DB_NAME "gline.db.new"

#define KEY_REASON "reason"
#define KEY_EXPIRES "expires"
#define KEY_ISSUER "issuer"

static heap_t gline_heap; /* key: expiry time, data: struct gline_entry* */
static dict_t gline_dict; /* key: target, data: struct gline_entry* */

static int
gline_comparator(const void *a, const void *b)
{
    const struct gline *ga=a, *gb=b;
    return ga->expires - gb->expires;
}

static void
free_gline_from_dict(void *data)
{
    struct gline *ent = data;
    free(ent->issuer);
    free(ent->target);
    free(ent->reason);
    free(ent);
}

static void
free_gline(struct gline *ent)
{
    dict_remove(gline_dict, ent->target);
}

static int
gline_for_p(void *key, void *data, void *extra)
{
    struct gline *ge = data;
    (void)key;
    return !strcasecmp(ge->target, extra);
}

static int
delete_gline_for_p(void *key, void *data, void *extra)
{
    struct gline *ge = data;

    (void)key;
    if (!strcasecmp(ge->target, extra)) {
        free_gline(ge);
        return 1;
    } else {
        return 0;
    }
}

static void
gline_expire(void *data)
{
    time_t stopped;
    void *wraa;

    (void)data;
    stopped = 0;
    while (heap_size(gline_heap)) {
        heap_peek(gline_heap, 0, &wraa);
        stopped = ((struct gline*)wraa)->expires;
        if (stopped > now) break;
        heap_pop(gline_heap);
        free_gline(wraa);
    }
    if (heap_size(gline_heap)) {
        timeq_add(stopped, gline_expire, NULL);
    }
}

int
gline_remove(const char *target, int announce)
{
    int res;
    res = dict_find(gline_dict, target, NULL) ? 1 : 0;
    if (heap_remove_pred(gline_heap, delete_gline_for_p, (char*)target)) {
        void *argh;
        struct gline *new_first;
        heap_peek(gline_heap, 0, &argh);
        if (argh) {
            new_first = argh;
            timeq_del(0, gline_expire, 0, TIMEQ_IGNORE_WHEN|TIMEQ_IGNORE_DATA);
            timeq_add(new_first->expires, gline_expire, 0);
        }
    }
    if (announce) irc_ungline(target);
    return res;
}

static void
gline_add_int(const char *issuer, const char *target, unsigned long duration, const char *reason, int announce)
{
    struct gline *ent;
    struct gline *prev_first;
    void *argh;

    heap_peek(gline_heap, 0, &argh);
    prev_first = argh;
    ent = dict_find(gline_dict, target, NULL);
    if (ent) {
        heap_remove_pred(gline_heap, gline_for_p, (char*)target);
        if (ent->expires < (time_t)(now + duration)) ent->expires = now + duration;
    } else {
        ent = malloc(sizeof(*ent));
        ent->issuer = strdup(issuer);
        ent->target = strdup(target);
        ent->expires = now + duration;
        ent->reason = strdup(reason);
        dict_insert(gline_dict, ent->target, ent);
    }
    heap_insert(gline_heap, ent, ent);
    if (!prev_first || (ent->expires < prev_first->expires)) {
	timeq_del(0, gline_expire, 0, TIMEQ_IGNORE_WHEN|TIMEQ_IGNORE_DATA);
	timeq_add(ent->expires, gline_expire, 0);
    }
    if (announce) irc_gline(ent->target, duration, ent->reason);
}

void
gline_add(const char *issuer, const char *target, unsigned long duration, const char *reason, int announce)
{
    gline_add_int(issuer, target, duration, reason, announce);
}

static int
gline_refresh_helper(void *key, void *data, void *extra)
{
    struct gline *ge = data;
    (void)key; (void)extra;
    irc_gline(ge->target, ge->expires-now, ge->reason);
    return 0;
}

void
gline_refresh_all(void)
{
    heap_remove_pred(gline_heap, gline_refresh_helper, 0);
}

unsigned int
gline_count(void)
{
    return dict_size(gline_dict);
}

static int
gline_add_record(const char *key, void *data, void *extra)
{
    struct record_data *rd = data;
    const char *issuer, *reason, *dstr;
    time_t expiration;

    (void)extra;
    reason = database_get_data(rd->d.object, KEY_REASON, RECDB_QSTRING);
    if (!reason) {
	log(MAIN_LOG, LOG_ERROR, "missing reason for gline of %s", key);
	return 0;
    }
    dstr = database_get_data(rd->d.object, KEY_EXPIRES, RECDB_QSTRING);
    if (!dstr) {
	log(MAIN_LOG, LOG_ERROR, "missing expiration for gline of %s", key);
	return 0;
    }
    expiration = (time_t)strtoul(dstr, NULL, 0);
    issuer = database_get_data(rd->d.object, KEY_ISSUER, RECDB_QSTRING);
    if (!issuer) issuer = "<unknown>";
    if (expiration > now) {
        gline_add_int(issuer, key, expiration - now, reason, 0);
    }
    return 0;
}

static void
gline_db_read(void)
{
    dict_t db;
    if ((db = parse_database(GLINE_DB_NAME))) {
        dict_foreach(db, gline_add_record, 0);
        free_database(db);
    }
}

static int
gline_insert_entry(void *key, void *data, void *extra)
{
    struct gline *ent = data;
    dict_t node = alloc_object();
    char expstr[12];
    (void)key;
    snprintf(expstr, sizeof(expstr), FMT_TIME_T, ent->expires);
    dict_insert(node, KEY_EXPIRES, alloc_record_data_qstring(expstr));
    dict_insert(node, KEY_REASON, alloc_record_data_qstring(ent->reason));
    dict_insert(node, KEY_ISSUER, alloc_record_data_qstring(ent->issuer));
    dict_insert(extra, ent->target, alloc_record_data_object(node));
    return 0;
}

static void
gline_db_write(void)
{
    FILE *dbfile;
    dict_t db = alloc_database();
    heap_remove_pred(gline_heap, gline_insert_entry, db);
    dbfile = fopen(TEMP_DB_NAME, "w");
    write_database(dbfile, db);
    fclose(dbfile);
    if (rename(TEMP_DB_NAME, GLINE_DB_NAME) == -1) {
        log(MAIN_LOG, LOG_ERROR, "unable to rename new gline db to old: %s\n", strerror(errno));
    }
    free_database(db);
}

static void
gline_db_cleanup(void)
{
    heap_delete(gline_heap);
    dict_delete(gline_dict);
}

void
gline_init(void)
{
    gline_heap = heap_new(gline_comparator);
    gline_dict = dict_new();
    dict_set_free_data(gline_dict, free_gline_from_dict);
    gline_db_read();
    reg_exit_func(gline_db_cleanup);
    reg_db_write_func(gline_db_write);
}

struct gline_discrim *
gline_discrim_create(struct userNode *user, struct userNode *src, unsigned int argc, unsigned char *argv[])
{
    unsigned int i;
    struct gline_discrim *discrim;

    discrim = malloc(sizeof(*discrim));
    memset(discrim, 0, sizeof(*discrim));
    discrim->limit = 50;

    for (i=0; i<argc; i++) {
        if (i > argc - 2) {
            send_message(user, src, MSG_MISSING_PARAMS, argv[i]);
            goto fail;
        }
        if (!strcasecmp(argv[i], "mask")) {
            if (!strcasecmp(argv[++i], "exact")) {
                discrim->target_mask_type = EXACT;
            } else if (!strcasecmp(argv[i], "subset")) {
                discrim->target_mask_type = SUBSET;
            } else if (!strcasecmp(argv[i], "superset")) {
                discrim->target_mask_type = SUPERSET;
            } else {
                discrim->target_mask_type = SUPERSET;
                i--;
            }
            if (i == argc - 1) {
                send_message(user, src, MSG_MISSING_PARAMS, argv[i]);
                goto fail;
            }
            if (!is_gline(argv[++i])) {
                send_message(user, src, MSG_INVALID_GLINE, argv[i]);
                goto fail;
            }
            discrim->target_mask = argv[i];
        } else if (!strcasecmp(argv[i], "limit")) {
            discrim->limit = strtoul(argv[++i], NULL, 0);
        } else if (!strcasecmp(argv[i], "reason")) {
            discrim->reason_mask = argv[++i];
        } else if (!strcasecmp(argv[i], "issuer")) {
            discrim->issuer_mask = argv[++i];
        } else if (!strcasecmp(argv[i], "after")) {
            discrim->min_expire = now + ParseInterval(argv[++i]);
        } else {
            send_message(user, src, MSG_INVALID_CRITERIA, argv[i]);
            goto fail;
        }
    }
    return discrim;
  fail:
    free(discrim);
    return NULL;
}

struct gline_search {
    struct gline_discrim *discrim;
    gline_search_func func;
    void *data;
    unsigned int hits;
};

static int
gline_discrim_match(struct gline *gline, struct gline_discrim *discrim)
{
    if ((discrim->issuer_mask && !match_ircglob(gline->issuer, discrim->issuer_mask))
        || (discrim->reason_mask && !match_ircglob(gline->reason, discrim->reason_mask))
        || (discrim->target_mask
            && (((discrim->target_mask_type == SUBSET)
                 && !match_ircglobs(discrim->target_mask, gline->target))
                || ((discrim->target_mask_type == EXACT)
                    && strcasecmp(discrim->target_mask, gline->target))
                || ((discrim->target_mask_type == SUPERSET)
                    && !match_ircglobs(gline->target, discrim->target_mask))))
        || (discrim->min_expire > gline->expires)) {
        return 0;
    }
    return 1;
}

static int
gline_search_helper(void *key, void *data, void *extra)
{
    struct gline *gline = data;
    struct gline_search *search = extra;

    (void)key;
    if (gline_discrim_match(gline, search->discrim)
        && (search->hits++ < search->discrim->limit)) {
        search->func(gline, search->data);
    }
    return 0;
}

unsigned int
gline_discrim_search(struct gline_discrim *discrim, gline_search_func gsf, void *data)
{
    struct gline_search search;
    search.discrim = discrim;
    search.func = gsf;
    search.data = data;
    search.hits = 0;
    heap_remove_pred(gline_heap, gline_search_helper, &search);
    return search.hits;
}
