/* helpfile.c - Help file loading and display
 * Copyright 2000-2002 srvx Development Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: helpfile.c,v 1.34.2.13.2.16 2002/05/06 13:20:53 entrope Exp $
 */

#include "config.h"
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>

#include "common.h"
#include "dict.h"
#include "hash.h"
#include "conf.h"
#include "log.h"
#include "nickserv.h"
#include "proto.h"
#include "recdb.h"

#include "helpfile.h"

#define DEFAULT_LINE_SIZE	MAX_LINE_SIZE
#define DEFAULT_TABLE_SIZE      80
#define HFMSG_HELP_NOT_STRING	"Help file error (help data was not a string)."

struct helpfile {
    dict_t db;
    expand_func_t expand;
};

extern struct userNode *global, *chanserv, *opserv, *nickserv;

void
table_send(struct userNode *from, const char *to, unsigned int size, irc_send_func irc_send, struct helpfile_table table) {
    unsigned int ii, jj, len, nreps, reps, tot_width, pos, spaces, *max_width;
    char line[MAX_LINE_SIZE+1];
    struct handle_info *hi = IsChannelName(to) ? NULL : GetUserN(to)->handle_info;

    /* If size or irc_send are 0, we should try to use a default. */
    if (!size) {
        if (hi) {
            if (!size) {
                if (hi->table_width) {
                    size = hi->table_width;
                } else if (hi->screen_width) {
                    size = hi->screen_width;
                } else {
                    size = DEFAULT_TABLE_SIZE;
                }
            }
        } else {
            if (!size) size = DEFAULT_TABLE_SIZE;
        }
        if (size > sizeof(line)) size = sizeof(line);
    }
    if (!irc_send) {
        if (hi) {
            irc_send = HANDLE_FLAGGED(hi, USE_PRIVMSG) ? irc_privmsg : irc_notice;
        } else {
            irc_send = IsChannelName(to) ? irc_privmsg : irc_notice;
        }
    }

    /* Figure out how wide columns should be */
    max_width = alloca(table.width * sizeof(int));
    for (jj=tot_width=0; jj<table.width; jj++) {
        /* Find the widest width for this column */
        max_width[jj] = 0;
        for (ii=0; ii<table.length; ii++) {
            len = strlen(table.contents[ii][jj]);
            if (len > max_width[jj]) max_width[jj] = len;
        }
        /* Separate columns with spaces */
        tot_width += max_width[jj] + 1;
    }
    /* How many rows to put in a line? */
    if ((table.flags & TABLE_REPEAT_ROWS) && (size > tot_width)) {
        nreps = size / tot_width;
    } else {
        nreps = 1;
    }
    /* Send headers line.. */
    if (table.flags & TABLE_NO_HEADERS) {
        ii = 0;
    } else {
        /* Sending headers needs special treatment: either show them
         * once, or repeat them as many times as we repeat the columns
         * in a row. */
        for (pos=ii=0; ii<((table.flags & TABLE_REPEAT_HEADERS)?nreps:1); ii++) {
            for (jj=0; jj<table.width; jj++) {
                len = strlen(table.contents[0][jj]);
                spaces = max_width[jj] - len;
                if (table.flags & TABLE_PAD_LEFT) while (spaces--) line[pos++] = ' ';
                memcpy(line+pos, table.contents[0][jj], len);
                pos += len;
                if (!(table.flags & TABLE_PAD_LEFT)) while (spaces--) line[pos++] = ' ';
                line[pos++] = ' ';
            }
        }
        line[--pos] = 0;
        irc_send(from->numeric, to, line);
        ii = 1;
    }
    /* Send the table. */
    for (pos=0, reps=0; ii<table.length; ) {
        for (jj=0; jj<table.width; jj++) {
            len = strlen(table.contents[ii][jj]);
            spaces = max_width[jj] - len;
            if (table.flags & TABLE_PAD_LEFT) while (spaces--) line[pos++] = ' ';
            memcpy(line+pos, table.contents[ii][jj], len);
            pos += len;
            if (!(table.flags & TABLE_PAD_LEFT)) while (spaces--) line[pos++] = ' ';
            line[pos++] = ' ';
        }
        ++ii; ++reps;
        if ((reps == nreps) || (ii==table.length)) {
            line[--pos] = 0;
            irc_send(from->numeric, to, line);
            pos = reps = 0;
        }
    }
    if (!(table.flags & TABLE_NO_FREE)) {
        /* Deallocate table memory (but not the string memory). */
        for (ii=0; ii<table.length; ii++) {
            free(table.contents[ii]);
        }
        free(table.contents);
    }
}

static int
vsend_message(const char *dest, struct userNode *src, struct handle_info *handle, int msg_type, expand_func_t expand_f, const char *format, va_list al)
{
    void (*irc_send)(const char *from, const char *to, const char *msg);
    char *input, line[MAX_LINE_SIZE];
    unsigned int size, pos, length, chars_sent;
    int ret, use_color;

    /* expand the parameters we got .. */
    chars_sent = 0;
    /* Normally, I'd pass NULL as the first parameter to vsnprintf.
     * But guess what?  It makes FreeBSD's libc allocate a buffer that it
     * never frees!  Surprise!
     * The kludge-around is to pass a non-NULL pointer instead.
     */
    if ((ret = vsnprintf(line, 0, format, al)) <= 0) {
        char *temp;
	/* non-C99 behavior */
	size = strlen(format);
	input = malloc(size);
	while ((ret = vsnprintf(input, size, format, al)) == -1) {
	    size <<= 1;
	    input = realloc(input, size);
	}
        temp = alloca(ret+1);
        strcpy(temp, input);
        free(input);
        input = temp;
    } else {
	/* C99 is nice about how vsnprintf fails */
	size = ret + 1;
	input = alloca(size);
	ret = vsnprintf(input, size, format, al);
    }

    /* figure out how to send the messages */
    if (handle) {
	msg_type |= (HANDLE_FLAGGED(handle, USE_PRIVMSG) ? 1 : 0);
	use_color = HANDLE_FLAGGED(handle, MIRC_COLOR);

	if (handle->screen_width) {
	    /* NickServ should clamp this value anyway. */
	    size = handle->screen_width;
	}
        if (size > sizeof(line)) size = sizeof(line);
    } else {
        size = sizeof(line);
        use_color = 1;
    }
    if (!size) size = DEFAULT_LINE_SIZE;
    switch (msg_type) {
        case 0:
            irc_send = irc_notice;
            break;
        case 2:
            irc_send = irc_wallchops;
            break;
        case 1:
        default:
            irc_send = irc_privmsg;
    }

    /* This used to be two passes, but if you do that and allow
     * arbitrary sizes for ${}-expansions (as with help indexes),
     * that requires a very big intermediate buffer.
     */
    pos = 0;
    while (*input) {
	char *value, *free_value, *min_backup;

        min_backup = input;
	while (*input && *input != '$' && *input != '\n' && pos < size) {
	    line[pos++] = *input++;
	}

	if (!*input) goto send_line;
        if (*input == '\n') {
            input++;
            goto send_line;
        }
	if (pos == size) {
            char *terminal;
            line[size - 1] = 0;
            terminal = strrchr(line, ' ');
            if (terminal) {
                int backup = pos - (terminal - line);
                /* min_backup is where we started -- after the last $-expression */
                if (input - backup < min_backup) {
                    backup = input - min_backup;
                    terminal = line + pos - backup;
                }
                *terminal = 0;
                input -= backup;
                pos -= backup;
            } else {
                /* Send the entire word, even though it's bigger than
                 * the requested maximum size.  But stop at line size.
                 */
                while (*input && (*input != ' ') && (pos < MAX_LINE_SIZE)) line[pos++] = *input++;
            }
            while (*input == ' ') input++;
	    goto send_line;
	}

        free_value = 0;
	switch (*(++input)) {
        /* Literal '$' or end of string. */
	case 0:
	    input--;
	case '$':
	    value = "$";
	    break;
	/* The following two expand to mIRC color codes if enabled
	   by the user. */
	case 'b':
	    value = use_color ? "\002" : "";
	    break;
	case 'o':
	    value = use_color ? "\017" : "";
	    break;
	case 'u':
	    value = use_color ? "\037" : "";
	    break;
	/* Service nicks. */
        case 'S':
            value = src->nick;
            break;
	case 'G':
	    value = global ? global->nick : "Global";
	    break;
	case 'C':
	    value = chanserv ? chanserv->nick : "ChanServ";
	    break;
	case 'O':
	    value = opserv ? opserv->nick : "OpServ";
	    break;
	case 'N':
            value = nickserv ? nickserv->nick : "NickServ";
            break;
        case 'n':
            if (nickserv) {
		const char *host = conf_server_name();

                length = strlen(nickserv->nick);
                value = alloca(length + strlen(host) + 2);
                strcpy(value, nickserv->nick);
                value[length++] = '@';
                strcpy(value + length, host);
            } else {
                value = "NickServ@Services";
            }
	    break;
	case 'H':
	    value = handle ? handle->handle : "Account";
	    break;
#define SEND_LINE() do { line[pos] = 0; if (pos > 0) irc_send(src->numeric, dest, line); chars_sent += pos; pos = 0; } while (0)
	/* Custom expansion handled by helpfile-specific function. */
	case '{':
	    if (expand_f) {
		char *name_end = input + 1;

		while (*name_end != '}' && *name_end) name_end++;
		if (*name_end) {
                    struct helpfile_expansion exp;
		    *name_end = 0;
		    exp = expand_f(input + 1);
                    switch (exp.type) {
                    case HF_STRING:
                        free_value = value = exp.value.str;
                        if (!value) value = "";
                        break;
                    case HF_TABLE:
                        /* Must send current line, then emit table. */
                        SEND_LINE();
                        table_send(src, dest, 0, irc_send, exp.value.table);
                        value = "";
                        break;
                    default:
                        value = "";
                        log(MAIN_LOG, LOG_ERROR, "Invalid exp.type %d from expansion function %p\n", exp.type, expand_f);
                        break;
                    }
		    input = name_end;
		    break;
		}
	    }

	/* Let it fall through when there's no expansion function or
	terminating '}'. */
	default:
		value = alloca(3);
		value[0] = '$';
		value[1] = *input;
		value[2] = 0;
	}
	input++;
        while ((pos + strlen(value) > size) || strchr(value, '\n')) {
            unsigned int avail;
            avail = size - pos - 1;
            length = strcspn(value, "\n ");
            if (length <= avail) {
                strncpy(line+pos, value, length);
                pos += length;
                value += length;
                /* copy over spaces, until (possible) end of line */
                while (*value == ' ') {
                    if (pos < size-1) line[pos++] = *value;
                    value++;
                }
            } else {
                /* word to send is too big to send now.. what to do? */
                if (pos > 0) {
                    /* try to put it on a separate line */
                    SEND_LINE();
                } else {
                    /* already at start of line; only send part of it */
                    strncpy(line, value, avail);
                    pos += avail;
                    value += length;
                    /* skip any trailing spaces */
                    while (*value == ' ') value++;
                }
            }
            /* if we're looking at a newline, send the accumulated text */
            if (*value == '\n') {
                SEND_LINE();
                value++;
            }
        }
        length = strlen(value);
	memcpy(line + pos, value, length);
        if (free_value) free(free_value);
	pos += length;
        if ((pos < size-1) && *input) continue;
      send_line:
        SEND_LINE();
#undef SEND_LINE
    }
    return chars_sent;
}

int
send_message(struct userNode *dest, struct userNode *src, const char *format, ...)
{
    int res;
    va_list ap;

    if (IsLocal(dest)) return 0;
    va_start(ap, format);
    res = vsend_message(dest->numeric, src, dest->handle_info, 0, NULL, format, ap);
    va_end(ap);
    return res;
}

int
send_target_message(int msg_type, const char *dest, struct userNode *src, const char *format, ...)
{
    int res;
    va_list ap;

    va_start(ap, format);
    res = vsend_message(dest, src, NULL, msg_type, NULL, format, ap);
    va_end(ap);
    return res;
}

int
_send_help(struct userNode *dest, struct userNode *src, expand_func_t expand, const char *format, ...)
{
    int res;

    va_list ap;
    va_start(ap, format);
    res = vsend_message(dest->numeric, src, dest->handle_info, 0, expand, format, ap);
    va_end(ap);
    return res;
}

int
send_help(struct userNode *dest, struct userNode *src, struct helpfile *hf, const char *topic)
{
    struct record_data *rec;
    const char *str;
    if (!topic) topic = "<index>";
    if (hf) {
	rec = dict_find(hf->db, topic, NULL);
    } else {
	rec = NULL;
    }
    if (!rec) {
	rec = dict_find(hf->db, "<missing>", NULL);
	if (!rec) {
	    _send_help(dest, src, hf->expand, MSG_TOPIC_UNKNOWN);
	    return 0;
	}
    }
    if (rec->type != RECDB_QSTRING) {
	_send_help(dest, src, hf->expand, HFMSG_HELP_NOT_STRING);
	return 0;
    }
    str = GET_RECORD_QSTRING(rec);
    return _send_help(dest, src, hf->expand, str);
}

int
unlistify_help(const char *key, void *data, void *extra)
{
    struct record_data *rd = data;
    dict_t newdb = extra;
    key = strdup(key);
    if (rd->type == RECDB_QSTRING) {
	dict_insert(newdb, key, alloc_record_data_qstring(GET_RECORD_QSTRING(rd)));
	return 0;
    } else if (rd->type == RECDB_STRING_LIST) {
	struct string_list *slist = GET_RECORD_STRING_LIST(rd);
	char *dest;
	unsigned int totlen, len, i;
	for (i=totlen=0; i<slist->used; i++) {
	    totlen = totlen + strlen(slist->list[i]) + 1;
	}
	dest = alloca(totlen+1);
	for (i=totlen=0; i<slist->used; i++) {
	    len = strlen(slist->list[i]);
	    memcpy(dest+totlen, slist->list[i], len);
	    dest[totlen+len] = '\n';
	    totlen = totlen + len + 1;
	}
	dest[totlen] = 0;
	dict_insert(newdb, key, alloc_record_data_qstring(dest));
	return 0;
    } else {
	return 1;
    }
}

struct helpfile *
open_helpfile(const char *fname, expand_func_t expand)
{
    struct helpfile *hf;
    dict_t db = parse_database(fname);
    hf = malloc(sizeof(struct helpfile));
    hf->db = alloc_database();
    dict_set_free_keys(hf->db, free);
    hf->expand = expand;
    if (db) {
	dict_foreach(db, unlistify_help, hf->db);
	free_database(db);
    }
    return hf;
}

void close_helpfile(struct helpfile *hf)
{
    free_database(hf->db);
    free(hf);
}
